# -*- coding: utf-8 -*-

# This work was created by participants in the DataONE project, and is
# jointly copyrighted by participating institutions in DataONE. For
# more information on DataONE, see our web site at http://dataone.org.
#
#   Copyright 2009-2016 DataONE
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Stdlib
import datetime
import logging
import urlparse

# 3rd party
import requests # pip install requests[security]
import requests.adapters
import cachecontrol # pip install cachecontrol
import requests_toolbelt # pip install requests-toolbelt
import requests_toolbelt.utils.dump

# D1
import d1_common.const
import d1_common.url
import d1_common.date_time

DEFAULT_NUMBER_OF_RETRIES = 3
DEFAULT_USE_CACHE = True
DEFAULT_USE_STREAM = False
DEFAULT_VERIFY_TLS = True


class Session(object):
  def __init__(
      self, base_url, cert_pem_path=None, cert_key_path=None, **kwargs_dict
  ):
    """The Session improves performance by keeping connection related state and
    allowing it to be reused in multiple API calls to a DataONE Coordinating
    Node or Member Node. This includes:

    - A connection pool
    - HTTP persistent connections (HTTP/1.1 and keep-alive)
    - Cached responses

    Based on Python Requests:
    - http://docs.python-requests.org/en/master/
    - http://docs.python-requests.org/en/master/user/advanced/#session-objects

    CacheControl is used for automated thread safe caching support:
    - http://cachecontrol.readthedocs.org/en/latest/

    :param base_url: DataONE Node REST service BaseURL.
    :type host: string

    :param cert_pem_path: Path to a PEM formatted certificate file.
    :type cert_pem_path: string

    :param cert_key_path: Path to a PEM formatted file that contains the private
      key for the certificate file. Only required if the certificate file does
      not itself contain the private key.
    :type cert_key_path: string

    :param timeout_sec: Time in seconds that requests will wait for a response.
    :type timeout_sec: float or int

    :param retries: Set number of times to try a request before failing. If not
      set, retries are still performed, using the default number of retries. To
      disable retries, set to 1.
    :type retries: int

    :param headers: headers that will be included with all connections.
    :type headers: dictionary

    :param query: URL query parameters that will be included with all
      connections.
    :type query: dictionary

    :param use_cache: Use the cachecontrol library to cache request responses.
      (default: True)
    :type use_cache: bool

    :param use_stream: Use streaming response. When enabled, responses must be
      completely read to free up the connection for reuse. (default:False)
    :type use_stream: bool

    :param verify_tls: Verify the server side TLS/SSL certificate.
      (default: True)
    :type verify_tls: bool

    :param user_agent: Override the default User-Agent string used by d1client.
    :type user_agent: str

    :param charset: Override the default Charset used by d1client.
      (default: UTF-8)
    :type charset: str

    :returns: None
    """
    self._base_url = base_url
    self._scheme, self._host, self._port, self._path = urlparse.urlparse(
      base_url
    )[:4]

    self._api_major = 1
    self._api_minor = 0

    # Adapter
    self._use_cache = kwargs_dict.pop('use_cache', DEFAULT_USE_CACHE)
    self._max_retries = kwargs_dict.pop('retries', DEFAULT_NUMBER_OF_RETRIES)

    # Default parameters for requests
    self._default_request_arg_dict = {
      'params':
        self._datetime_to_iso8601(kwargs_dict.pop('query', {})),
      'timeout':
        kwargs_dict.pop('timeout_sec', d1_common.const.RESPONSE_TIMEOUT),
      'stream':
        kwargs_dict.pop('use_stream', DEFAULT_USE_STREAM),
      'verify':
        kwargs_dict.pop('verify_tls', DEFAULT_VERIFY_TLS),
      'headers': {
        'User-Agent':
          kwargs_dict.pop('user_agent', d1_common.const.USER_AGENT),
        'Charset':
          kwargs_dict.pop('charset', d1_common.const.DEFAULT_CHARSET),
      },
    }
    self._default_request_arg_dict['headers'
                                   ].update(kwargs_dict.pop('headers', {}))
    self._default_request_arg_dict.update(kwargs_dict)
    # Requests wants cert path as string if single file and tuple if cert/key
    # pair.
    if cert_pem_path is not None:
      self._default_request_arg_dict['cert'] = (
        cert_pem_path, cert_key_path
        if cert_key_path is not None else cert_pem_path
      )

    self._session = self._create_requests_session()

  @property
  def base_url(self):
    return self._base_url

  def GET(self, rest_path_list, **kwargs):
    """Send a GET request.
    See requests.sessions.request for optional parameters.
    :returns: Response object
    """
    return self._request('GET', rest_path_list, **kwargs)

  def HEAD(self, rest_path_list, **kwargs):
    """Send a HEAD request.
    See requests.sessions.request for optional parameters.
    :returns: Response object
    """
    kwargs.setdefault('allow_redirects', False)
    return self._request('HEAD', rest_path_list, **kwargs)

  def POST(self, rest_path_list, **kwargs):
    """Send a POST request with optional streaming multipart encoding.
    See requests.sessions.request for optional parameters. To post regular data,
    pass a string, iterator or generator as the {data} argument. To post a
    multipart stream, pass a dictionary of multipart elements as the
    {fields} argument. E.g.:

    fields = {
      'field0': 'value',
      'field1': 'value',
      'field2': ('filename.xml', open('file.xml', 'rb'), 'application/xml')
    }

    :returns: Response object
    """
    fields = kwargs.pop('fields', None)
    if fields is not None:
      return self._send_mmp_stream('POST', rest_path_list, fields, **kwargs)
    else:
      return self._request('POST', rest_path_list, **kwargs)

  def PUT(self, rest_path_list, **kwargs):
    """Send a PUT request with optional streaming multipart encoding.
    See requests.sessions.request for optional parameters. See post() for
    parameters.
    :returns: Response object
    """
    fields = kwargs.pop('fields', None)
    if fields is not None:
      return self._send_mmp_stream('PUT', rest_path_list, fields, **kwargs)
    else:
      return self._request('PUT', rest_path_list, **kwargs)

  def DELETE(self, rest_path_list, **kwargs):
    """Send a DELETE request.
    See requests.sessions.request for optional parameters.
    :returns: Response object
    """
    return self._request('DELETE', rest_path_list, **kwargs)

  def get_curl_command_line(self, method, url, **kwargs):
    """Get request as cURL command line for debugging.
    """
    if kwargs.get('query'):
      url = u'{0}?{1}'.format(url, d1_common.url.urlencode(kwargs['query']))
    curl_cmd = [u'curl -X {0}'.format(method)]
    for k, v in kwargs['headers'].items():
      curl_cmd.append(u'-H "{0}: {1}"'.format(k, v))
    curl_cmd.append(u'{0}'.format(url))
    return ' '.join(curl_cmd)

  #
  # Private
  #

  def _create_requests_session(self):
    session = requests.Session()
    if self._use_cache:
      adapter_cls = cachecontrol.CacheControlAdapter
    else:
      adapter_cls = requests.adapters.HTTPAdapter
    adapter = adapter_cls(max_retries=self._max_retries)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session

  def _send_mmp_stream(self, method, rest_path_list, fields, **kwargs):
    url = self._prep_url(rest_path_list)
    kwargs = self._prep_request_kwargs(kwargs)
    mmp_stream = requests_toolbelt.MultipartEncoder(fields=fields)
    kwargs['data'] = mmp_stream
    kwargs['headers'].update({
      'Content-Type': mmp_stream.content_type,
      'Content-Length': str(mmp_stream.len),
      # 'Connection': 'close', # ###############################
    })
    return self._session.request(method, url, **kwargs)

  def _request(self, method, rest_path_list, **kwargs):
    url = self._prep_url(rest_path_list)
    kwargs = self._prep_request_kwargs(kwargs)
    logging.debug(self.get_curl_command_line(method, url, **kwargs))
    return self._session.request(method, url, **kwargs)
    # return self._session.request(method, url, **kwargs)
    # return self._create_requests_session().request(method, url, **kwargs)
    # with requests.Session() as s:
    #   s.keep_alive = False
    #   # print kwargs
    #   kwargs['headers']['Connection'] = 'close'
    #   return s.request(method, url, **kwargs)

  def _prep_url(self, rest_path_list):
    if isinstance(rest_path_list, basestring):
      rest_path_list = [rest_path_list]
    return d1_common.url.joinPathElements(
      self._base_url,
      self._get_api_version_path_element(),
      *self._encode_path_elements(rest_path_list)
    )

  def _prep_request_kwargs(self, kwargs_in_dict):
    # kwargs that require translation
    kwargs_dict = {
      'timeout': kwargs_in_dict.pop('timeout_sec', None),
      'stream': kwargs_in_dict.pop('use_stream', None),
      'verify': kwargs_in_dict.pop('verify_tls', None),
      'params': self._datetime_to_iso8601(kwargs_in_dict.pop('query', {})),
    }
    kwargs_dict.update(kwargs_in_dict)
    kwargs_dict = self._remove_none_value_items(kwargs_dict)
    result_dict = self._default_request_arg_dict.copy()
    result_dict.update(kwargs_dict)
    return result_dict

  def _datetime_to_iso8601(self, query_dict):
    """Encode any datetime query parameters to ISO8601."""
    return {
      k: v if not isinstance(v, datetime.datetime) else v.isoformat()
      for k, v in query_dict.items()
    }

  def _remove_none_value_items(self, d):
    return {k: v for k, v in d.items() if v is not None}

  def _encode_path_elements(self, path_element_list):
    return [d1_common.url.encodePathElement(v) for v in path_element_list]

  def _get_api_version_path_element(self):
    return 'v{}'.format(self._api_major)

  def _get_api_version_xml_type(self):
    if (self._api_major, self._api_minor) == (1, 0):
      return 'v1'
    else:
      return 'v{}.{}'.format(self._api_major, self._api_minor)

  def _get_expected_schema_type_attribute(self):
    return '{}{}'.format(
      d1_common.const.DATAONE_SCHEMA_ATTRIBUTE_BASE,
      self._get_api_version_xml_type()
    )
